/*
 * $Id: ExportPinPixelsDialog.java,v 1.9 2005/09/02 20:10:40 max Exp $
 *
 * Copyright (C) 2002 by Brockmann Consult (info@brockmann-consult.de)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation. This program is distributed in the hope it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.bc.beam.visat;

import java.awt.GridBagConstraints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.BorderFactory;
import javax.swing.ButtonGroup;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JTextField;

import org.esa.beam.framework.datamodel.Product;
import org.esa.beam.framework.param.ParamChangeEvent;
import org.esa.beam.framework.param.ParamChangeListener;
import org.esa.beam.framework.param.Parameter;
import org.esa.beam.framework.ui.GridBagUtils;
import org.esa.beam.framework.ui.ModalDialog;
import org.esa.beam.framework.ui.product.ProductExpressionPane;
import org.esa.beam.util.Debug;
import org.esa.beam.visat.VisatApp;

/**
 * This class defines a modal dialog which lets the user specify a quadratic region around a
 * selected pin and an arithmetic expression.
 * 
 * @author Maximilian Aulinger
 */

public class ExportPinPixelsDialog extends ModalDialog {

    private static final String _PARAM_NAME_USE_ARTIHMETIC_EXPRESSION = "useArithmeticExpr";
    private static final int _MAX_REGION_SIZE = 25;

    private final VisatApp _visatApp;
    private final Product _product;

    private Parameter _paramRegion;
    private Parameter _paramExpression;
    private Parameter _paramUseExpression;
    private JRadioButton _buttonExportAllPins;
    private JRadioButton _buttonExportSelectedPin;
    private JRadioButton _buttonUseExpressionAsFilter;
    private JRadioButton _buttonUseExpressionAsMarker;

    private JButton _editExpressionButton;
    private JTextField _pinLabelTextField;

    /**
     * Initialises a new instance of this class.
     * 
     * @param visatApp a reference to the VISAT application instance
     * @param product the product providing the pixel values
     */
    ExportPinPixelsDialog(final VisatApp visatApp, final Product product) {
        super(visatApp.getMainFrame(), "Export Pin Pixels", ModalDialog.ID_OK_CANCEL, null);
        _product = product;
        _visatApp = visatApp;
        initParameter();
        createUI();
    }

    /**
     * <code>true</code> if the user has selected "Use arithmetic expression".
     */
    boolean isUseExpression() {
        return ((Boolean) _paramUseExpression.getValue()).booleanValue();
    }

    /**
     * <code>true</code> if the user has selected "Export selected pin".
     */
    boolean isExportSelectedPinOnly() {
        return _buttonExportSelectedPin.isSelected();
    }

    /**
     * <code>true</code> if the user has selected "Use expression as filter".
     */
    boolean isUseExpressionAsFilter() {
        return _buttonUseExpressionAsFilter.isSelected();
    }

    /**
     * Returns the region's size specified by the user.
     */
    String getRegion() {
        return _paramRegion.getValueAsText();
    }

    /**
     * Returns the region's width/height
     */
    int getRegionSize() {
        return extractRegionSize(getRegion());
    }

    /**
     * Returns the arithmetic expression specified by the user.
     */
    String getExpression() {
        final String expression = _paramExpression.getValueAsText();
        if (expression != null && expression.trim().length() > 0) {
            return expression;
        } else {
            return null;
        }
    }

    /**
     * Initialises the selectable parameters.
     */
    private void initParameter() {
        final ParamChangeListener paramChangeListener = createParamChangeListener();

        final String[] possibleRegions = defineRegionSelection(_MAX_REGION_SIZE);

        _paramRegion = new Parameter("exportRegion", possibleRegions[0]);
        _paramRegion.getProperties().setLabel("Region size:"); /* I18N */
        _paramRegion.getProperties().setValueSet(possibleRegions);
        _paramRegion.getProperties().setValueSetBound(true);

        _paramExpression = new Parameter("arithmeticExpr", "");
        _paramExpression.getProperties().setLabel("Expression"); /* I18N */
        _paramExpression.getProperties().setDescription("Arithmetic expression"); /* I18N */
        _paramExpression.getProperties().setNumRows(3);
        _paramExpression.setUIEnabled(false);

        _paramUseExpression = new Parameter(_PARAM_NAME_USE_ARTIHMETIC_EXPRESSION, new Boolean(false));
        _paramUseExpression.getProperties().setLabel("Use expression"); /* I18N */
        _paramUseExpression.addParamChangeListener(paramChangeListener);
    }

    /**
     * Creates the UI elements.
     */
    private void createUI() {

        final JPanel panel = GridBagUtils.createPanel();
        int line = 0;
        final GridBagConstraints gbc = new GridBagConstraints();

        gbc.gridy = ++line;
        GridBagUtils.addToPanel(panel, createPinPane(), gbc,
                "weighty=0, weightx=1, fill=BOTH, gridwidth=3, anchor=NORTHWEST");

        gbc.gridy = ++line;
        GridBagUtils.addToPanel(panel, createRegionPane(), gbc,
                "weighty=0, weightx=1, fill=BOTH, gridwidth=3, anchor=NORTHWEST");

        gbc.gridy = ++line;
        GridBagUtils.addToPanel(panel, createExpressionPane(), gbc,
                "weighty=1, weightx=1, fill=BOTH, gridwidth=3, anchor=NORTHWEST");

        setContent(panel);
    }

    private JPanel createPinPane() {
        final JPanel pinPane = GridBagUtils.createPanel();
        int line = 0;
        final GridBagConstraints gbc = new GridBagConstraints();

        gbc.gridy = ++line;
        GridBagUtils.addToPanel(pinPane, new JLabel("Selected pin: "), gbc,
                "weightx=1, fill=HORIZONTAL, gridwidth=3, anchor=NORTHWEST");
        gbc.gridy = ++line;
        _pinLabelTextField = new JTextField("");
        _pinLabelTextField.setEditable(false);
        GridBagUtils.addToPanel(pinPane, _pinLabelTextField, gbc,
                "fill=HORIZONTAL, gridwidth=3, anchor=NORTHWEST");

        _buttonExportAllPins = new JRadioButton("Export all pins"); /* I18N */
        _buttonExportAllPins.setSelected(true);
        _buttonExportSelectedPin = new JRadioButton("Export selected pin"); /* I18N */

        final ButtonGroup pinsGroup = new ButtonGroup();
        pinsGroup.add(_buttonExportAllPins);
        pinsGroup.add(_buttonExportSelectedPin);

        gbc.gridy = ++line;
        GridBagUtils.addToPanel(pinPane, _buttonExportAllPins, gbc,
                "weighty=0, weightx=0, fill=NONE, gridwidth=1, anchor=WEST");
        GridBagUtils.addToPanel(pinPane, _buttonExportSelectedPin, gbc,
                "weighty=0, weightx=0, fill=NONE, gridwidth=1, anchor=WEST");

        return pinPane;
    }

    private JPanel createRegionPane() {
        final JPanel regionPane = GridBagUtils.createPanel();
        int line = 0;
        final GridBagConstraints gbc = new GridBagConstraints();

        gbc.gridy = ++line;
        GridBagUtils.addToPanel(regionPane, _paramRegion.getEditor().getLabelComponent(), gbc,
                "weightx=1, gridwidth=3, fill=BOTH, anchor=WEST");
        gbc.gridy = ++line;
        GridBagUtils.addToPanel(regionPane, _paramRegion.getEditor().getComponent(), gbc,
                "gridwidth=3, fill=BOTH, anchor=WEST");

        return regionPane;
    }

    private JPanel createExpressionPane() {

        final JPanel expressionPane = GridBagUtils.createPanel();
        int line = 0;
        final GridBagConstraints gbc = new GridBagConstraints();
        expressionPane.setBorder(BorderFactory.createTitledBorder("Arithmetic Expression"));

        gbc.gridy = ++line;
        GridBagUtils.addToPanel(expressionPane, _paramUseExpression.getEditor().getComponent(), gbc,
                "weighty=0, weightx=0, fill=NONE, gridwidth=0, anchor=NORTHWEST");

        gbc.gridy = ++line;
        GridBagUtils.addToPanel(expressionPane, _paramExpression.getEditor().getLabelComponent(), gbc,
                "fill=HORIZONTAL, gridwidth=3, anchor=NORTHWEST");
        gbc.gridy = ++line;
        GridBagUtils.addToPanel(expressionPane, _paramExpression.getEditor().getComponent(), gbc,
                "weighty=1,weightx=1 , fill=BOTH, gridwidth=3, anchor=NORTHWEST");

        _editExpressionButton = new JButton("Edit Expression...");
        _editExpressionButton.addActionListener(createEditExpressionButtonListener());
        _editExpressionButton.setEnabled(false);

        gbc.gridy = ++line;
        GridBagUtils.addToPanel(expressionPane, _editExpressionButton, gbc,
                "weighty=0,weightx=0, fill=NONE, gridwidth=3, anchor=NORTHEAST");

        _buttonUseExpressionAsFilter = new JRadioButton("Use expression as filter"); /* I18N */
        _buttonUseExpressionAsFilter.setEnabled(false);
        
        _buttonUseExpressionAsMarker = new JRadioButton("Add column with pixel relevance information"); /* I18N */
        _buttonUseExpressionAsMarker.setEnabled(false);
        _buttonUseExpressionAsMarker.setSelected(true);

        final ButtonGroup expressionGroup = new ButtonGroup();
        expressionGroup.add(_buttonUseExpressionAsFilter);
        expressionGroup.add(_buttonUseExpressionAsMarker);

        gbc.gridy = ++line;
        GridBagUtils.addToPanel(expressionPane, _buttonUseExpressionAsFilter, gbc,
                "weighty=0, weightx=0, fill=NONE, gridwidth=1, anchor=WEST");
        gbc.gridy = ++line;
        GridBagUtils.addToPanel(expressionPane, _buttonUseExpressionAsMarker, gbc,
                "weighty=0, weightx=0, fill=NONE, gridwidth=1, anchor=WEST");

        return expressionPane;
    }

    /**
     * Updates the UI elements.
     */
    private void updateUIState() {
        final boolean useExpression = ((Boolean) _paramUseExpression.getValue()).booleanValue();
        _paramExpression.setUIEnabled(useExpression);
        _editExpressionButton.setEnabled(useExpression);
        _buttonUseExpressionAsFilter.setEnabled(useExpression);
        _buttonUseExpressionAsMarker.setEnabled(useExpression);
    }

    protected boolean verifyUserInput() {
        final String expression = _paramExpression.getValueAsText();
        final boolean useExpression = ((Boolean) _paramUseExpression.getValue()).booleanValue();
        if (useExpression && !isValidExpression(expression)) {
            VisatApp.getApp().showErrorDialog("Please check the expression you have entered.\nIt is not valid."); /* I18N */
            return false;
        }
        return true;
    }

    /**
     * Shows the ExportPinPixels dialog window.
     * 
     * @param pinLabel the label of the selected pin
     * @return the int value associated with the button the user has pressed. See org.esa.beam.framework.ui.ModalDialog
     */
    int show(final String pinLabel) {
        _pinLabelTextField.setText(pinLabel);
        boolean onlyOnePin = (_product.getNumPins() == 1);
        _buttonExportSelectedPin.setSelected(onlyOnePin);
        _buttonExportAllPins.setEnabled(!onlyOnePin);
        return super.show();
    }

    /**
     * @deprecated use {@link #show(String)} instead.
     */
    public int show() {
        return super.show();
    }

    /**
     * Extracts an integer value "size" out of a String size x size
     */
    private int extractRegionSize(final String userSelectedRegionSize) {
        final String[] extractedRegion = userSelectedRegionSize.split("x");
        int extractedRegionSize = new Integer(extractedRegion[0].trim()).intValue();
        return extractedRegionSize;
    }

    /**
     * Defines an array with all possible region sizes (used as pulldown menu).
     * 
     * @param maxRegionSize the maximum height/width for the quadratic region around the pin
     * 
     * @return the array with the possible options
     */
    private String[] defineRegionSelection(final int maxRegionSize) {
        if (maxRegionSize == 0) {
            return null;
        }
        final String[] regionSelection = new String[(maxRegionSize - 1) / 2];
        for (int i = 0; (3 + i * 2 <= maxRegionSize); i++) {
            final int n = (3 + i * 2);
            regionSelection[i] = (n + " x " + n + " pixels");
        }
        return regionSelection;
    }

    /**
     * Checks the arithmetic expression.
     */
    private boolean isValidExpression(final String expression) {
        if (expression == null || expression.length() == 0) {
            return false;
        }
        return _product.isCompatibleBandArithmeticExpression(expression);
    }

    /**
     * Defines the ActionListener for the "Edit Expression"-button.
     */
    private ActionListener createEditExpressionButtonListener() {
        return new ActionListener() {

            public void actionPerformed(final ActionEvent e) {
                final Product[] products = { _product };
                ProductExpressionPane pep = ProductExpressionPane.createBooleanExpressionPane(products, _product,
                        _visatApp.getPreferences());
                pep.setCode(_paramExpression.getValueAsText());
                final int status = pep.showModalDialog(getJDialog(), "Arithmetic Expression Editor"); /* I18N */
                if (status == ModalDialog.ID_OK) {
                    _paramExpression.setValue(pep.getCode(), null);
                    Debug.trace("BandArithmetikDialog: expression is: " + pep.getCode()); /* I18N */
                }
                pep.dispose();
                pep = null;
            }
        };
    }

    /**
     * The ParameterChangeListener for the useExpression parameter.
     */
    private ParamChangeListener createParamChangeListener() {
        return new ParamChangeListener() {

            public void parameterValueChanged(final ParamChangeEvent event) {
                updateUIState();
            }
        };
    }
}
