/*
 * $Id$
 *
 * Copyright (C) 2002 by Brockmann Consult (info@brockmann-consult.de)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation. This program is distributed in the hope it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.bc.beam.visat;

import java.awt.Point;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.StringWriter;
import java.io.Writer;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map.Entry;

import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;

import org.esa.beam.framework.datamodel.Pin;
import org.esa.beam.framework.datamodel.Product;
import org.esa.beam.framework.ui.ModalDialog;
import org.esa.beam.framework.ui.SelectExportMethodDialog;
import org.esa.beam.framework.ui.SwingWorker;
import org.esa.beam.framework.ui.UIUtils;
import org.esa.beam.framework.ui.command.Command;
import org.esa.beam.framework.ui.command.CommandEvent;
import org.esa.beam.framework.ui.command.CommandListener;
import org.esa.beam.framework.ui.command.CommandManager;
import org.esa.beam.framework.ui.product.ProductSceneView;
import org.esa.beam.util.SystemUtils;
import org.esa.beam.util.io.FileUtils;
import org.esa.beam.visat.VisatApp;
import org.esa.beam.visat.VisatPlugIn;

import com.bc.jexp.ParseException;
import com.bc.swing.ProgressMonitor;

/**
 * Installs a plug-in wich can be used to export all pixel values within a quadric region around one
 * or all pin(s) in a product.
 * 
 * @author Maximilian Aulinger
 * @version 1.3
 */

public class ExportPinPixelsVPI implements VisatPlugIn {

    private static final String ERR_MSG_BASE = "Pin pixels cannot be exported:\n";
    private static final String COMMAND_ID = "exportPinPixels";
    private static final String COMMAND_NAME = "Export Pin Pixels";

    private VisatApp _visatApp;
    private ExportPinPixelsDialog _dialog;
    private Product _product;

    public void initPlugIn(final VisatApp visatApp) {
        _visatApp = visatApp;

        // Create a new VISAT command handler (as instance of an anonymous class)
        // that lets user export the relevant pixels.
        final CommandListener commandListener = new CommandListener() {

            /**
             * Invoked when a command action is performed.
             * 
             * @param event the command event
             */
            public void actionPerformed(final CommandEvent event) {
                exportPinPixels(event);
            }

            /**
             * Called when a command should update its state. <p/> <p/> This method can contain some
             * code which analyzes the underlying element and makes a decision whether this item or
             * group should be made visible/invisible or enabled/disabled etc.
             * 
             * @param event the command event
             */
            public void updateState(final CommandEvent event) {
                boolean enabled = false;
                final Product product = getSelectedProduct();
                if (product != null) {
                    enabled = product.getNumPins() > 0;
                }
                event.getCommand().setEnabled(enabled);
            }
        };

        // Now create the actual command with the command handler we've just created above.
        // We use some of the multiple command property setters here to configure our command.

        final CommandManager commandManager = visatApp.getCommandManager();
        final Command command = commandManager.createExecCommand(COMMAND_ID, commandListener);
        command.setEnabled(false); // disable command
        command.setText(COMMAND_NAME + "..."); // e.g. menu item text
        command.setShortDescription(COMMAND_NAME); // e.g. text used as tooltip
        // text
        command.setLongDescription(COMMAND_NAME); // e.g. text shown in status
        // bar
        command.setParent("export"); // Place in "export" command group
        command.setContexts(new String[] { "image" }); // Show command, if
        // context is "image"
        // (context menu)

    }

    public void updateComponentTreeUI() {
        if (_dialog != null) {
            SwingUtilities.updateComponentTreeUI(_dialog.getJDialog());
        }
    }

    /**
     * Performs the "Export Pin Pixels" command.
     */
    private void exportPinPixels(final CommandEvent event) {
        _product = getSelectedProduct();

        ensureThatAPinIsSelected(_product);
        final Pin selectedPin = _product.getSelectedPin();
        if (_dialog == null) {
            _dialog = new ExportPinPixelsDialog(_visatApp, _product);
        }

        // shows a dialog which lets the user specify the region he wants to export.
        final int dialogAnswer = _dialog.show(selectedPin.getLabel());

        // return if user pressed the Cancel-button
        if (ModalDialog.ID_OK != dialogAnswer) {
            return;
        }

        final HashMap pinPixels = assignControllParameters();
        if (getNumPixelsToExport(pinPixels) == 0) {
            _visatApp.showErrorDialog("There are no pixels to export."); /* I18N */
            return;
        }

        final Writer out = getOutputWriter(pinPixels, event, _product);

        if (out == null) {
            return; // Export Pin Pixels was canceled by the export method dialog
        }

        final TabSeparatedPinPixelsWriter regionWriter = new TabSeparatedPinPixelsWriter(out, _product.getBands(),
                _product.getGeoCoding());

        final SwingWorker swingWorker = createWorkerInstance(pinPixels, regionWriter, out);

        // show wait-cursor
        UIUtils.setRootFrameWaitCursor(_visatApp.getMainFrame());
        // show message in status bar
        _visatApp.setStatusBarMessage("Exporting Pin pixels..."); /* I18N */

        // Start separate worker thread.
        swingWorker.start();
    }

    /**
     * Ensures that a pin is selected in the given product.
     */
    private void ensureThatAPinIsSelected(final Product product) {
        if (product.getSelectedPin() == null) {
            product.getPinAt(0).setSelected(true);
        }
    }

    /**
     * Gets a target Writer for the export data.
     */
    private Writer getOutputWriter(final HashMap pinPixels, final CommandEvent event, final Product product) {
        // Get export method from user
        final int method = getExportMethod(getNumPixelsToExport(pinPixels), event);

        if (method == SelectExportMethodDialog.EXPORT_TO_CLIPBOARD) {
            // Write into string buffer
            return new StringWriter();
        } else if (method == SelectExportMethodDialog.EXPORT_TO_FILE) {
            // Write into file, get file from user
            final File file = promptForFile(_visatApp, createDefaultFileName(product));
            if (file == null) {
                return null; // Cancel
            }
            try {
                return new FileWriter(file);
            } catch (IOException e) {
                _visatApp.showErrorDialog(COMMAND_NAME, ERR_MSG_BASE + "Failed to create file '" + file + "':\n"
                        + e.getMessage()); /* I18N */
                return null; // Error
            }
        } else {
            return null; // Cancel
        }
    }

    /**
     * Opens a modal dialog that asks the user which method to use in order to export the pin
     * pixels.
     */
    private int getExportMethod(final int numPixels, CommandEvent event) {
        int result;

        final String questionText = "How do you want to export the pixel values?\n"; /* I18N */

        final String numPixelsText;
        if (numPixels == 1) {
            numPixelsText = "One pixel will be exported.\n"; /* I18N */
        } else {
            numPixelsText = numPixels + " pixels will be exported.\n"; /* I18N */
        }
        result = SelectExportMethodDialog.run(_visatApp.getMainFrame(), getWindowTitle(), questionText + numPixelsText,
                event.getCommand().getHelpId());

        return result;
    }

    /**
     * Opens a modal file chooser dialog that prompts the user to select the output file name.
     * 
     * @param visatApp the VISAT application
     * 
     * @return the selected file, <code>null</code> means "Cancel"
     */
    private File promptForFile(final VisatApp visatApp, final String defaultFileName) {
        // Loop while the user does not want to overwrite a selected, existing file or if the user
        // presses "Cancel"

        File file = null;
        while (file == null) {
            file = visatApp.showFileSaveDialog(COMMAND_NAME, false, null, ".txt", defaultFileName, COMMAND_ID
                    + ".lastDir");
            if (file == null) {
                return null; // Cancel
            } else if (file.exists()) {
                final int status = JOptionPane.showConfirmDialog(visatApp.getMainFrame(), "The file '" + file
                        + "' already exists.\n" + /* I18N */
                        "Overwrite it?", /* I18N */
                "VISAT - " + COMMAND_NAME, JOptionPane.YES_NO_CANCEL_OPTION, JOptionPane.WARNING_MESSAGE);
                if (status == JOptionPane.CANCEL_OPTION) {
                    return null; // Cancel
                } else if (status == JOptionPane.NO_OPTION) {
                    file = null; // No, do not overwrite, let user select
                    // other file
                }
            }
        }
        return file;
    }

    /**
     * Creates the export Data. Can be <code>null</code> if an error occures while reading the
     * product data or user input.
     */
    private HashMap assignControllParameters() {

        final Pin[] exportPins;
        if (_dialog.isExportSelectedPinOnly()) {
            exportPins = new Pin[] { _product.getSelectedPin() };
        } else {
            exportPins = _product.getPins();
        }

        final int regionSize = _dialog.getRegionSize();
        String expression = null;
        boolean useExpressionAsFilter = true;

        if (_dialog.isUseExpression()) {
            expression = _dialog.getExpression();
            useExpressionAsFilter = _dialog.isUseExpressionAsFilter();
        }

        HashMap exportData = null;

        try {
            exportData = generateOutputData(regionSize, expression, exportPins, useExpressionAsFilter);
        } catch (IOException e) {
            VisatApp.getApp().showErrorDialog("An I/O error occured:\n" + e.getMessage()); /* I18N */
            return null;
        } catch (ParseException e) {
            VisatApp.getApp().showErrorDialog("Please check the expression you have entered.\nIt is not valid."); /* I18N */
            return null;
        }

        return exportData;
    }

    private SwingWorker createWorkerInstance(final HashMap regions, final TabSeparatedPinPixelsWriter regionWriter,
            final Writer out) {
        // Create a progress monitor and adds them to the progress controller pool in order to show
        // export progress
        final ProgressMonitor monitor = new ProgressMonitor(_visatApp.getMainFrame(), COMMAND_NAME);

        // Create a new swing worker instance (as instance of an anonymous class).
        // When the swing worker's start() method is called, a new separate thread is started.
        // The swing worker's construct() method is executed in that thread, so that VISAT can keep
        // on handling user events.
        final SwingWorker swingWorker = new SwingWorker(monitor) {

            /**
             * Compute the value to be returned by the <code>get</code> method. This method is
             * executed in a separate worker thread.
             */
            public Object construct() {
                try {
                    final boolean success = regionWriter.writePinPixels(_dialog.getRegionSize(), _dialog
                            .getExpression(), regions);
                    if (success) {
                        maybeCopyToClipboard(out);
                    }
                } catch (IOException e) {
                    return e;
                }
                return null;
            }

            /**
             * Called on the event dispatching thread (not on the worker thread) after the
             * <code>construct</code> method has returned.
             */
            public void finished() {
                // clear status bar
                _visatApp.clearStatusBarMessage();
                // show default-cursor
                UIUtils.setRootFrameDefaultCursor(_visatApp.getMainFrame());
                // On error, show error message
                if (getValue() instanceof Exception) {
                    final Exception e = (Exception) getValue();
                    _visatApp.showErrorDialog(COMMAND_NAME, ERR_MSG_BASE + e.getMessage());
                }
            }
        };
        return swingWorker;
    }

    private void maybeCopyToClipboard(final Writer out) {
        if (out instanceof StringWriter) {
            final StringWriter stringWriter = (StringWriter) out;
            SystemUtils.copyToClipboard(stringWriter.toString());
        }
    }

    /**
     * Returns the product the user has selected in beam. Can be <code>null</code>.
     */
    private Product getSelectedProduct() {
        final ProductSceneView view = _visatApp.getSelectedProductSceneView();
        if (view != null) {
            return view.getProduct();
        }
        return null;
    }

    /**
     * Generates the output data as a hashmap. For each pin in the <code>pins</code> array the relevant pixel
     * coordinates around it are stored as a point array in a hashmap. The respective pin is taken
     * as key to access the pixel coordinates.
     * 
     * @param size the size of the pixel region to be exported
     * @param expression an arithmetic expression for classifying pixel relevance
     * @param pins the pins defining the center of the pixels regions
     * @param useExpressionAsFilter toggle between filter pixels or mark relevance in an additional column
     * @return a hashmap providing the pixel coordinates
     * 
     * @throws IOException
     * @throws ParseException
     */
    private HashMap generateOutputData(final int size,final String expression,final Pin[] pins,final boolean useExpressionAsFilter) throws IOException, ParseException {
        HashMap<Pin,Object[]> outputData = new HashMap<Pin,Object[]>();
        for (int i = 0; i < pins.length; i++) {
            Point actualPin = new Point((int) pins[i].getPixelPos().x, (int) pins[i].getPixelPos().y);
            PinPixelsGenerator pinPixels = new PinPixelsGenerator(_product);
            if (useExpressionAsFilter) {
                Point[] pixels = pinPixels.generateQuadricPixelRegion(actualPin, size, expression);
                outputData.put(pins[i], pixels);
            } else {
                Point[] pixels = pinPixels.generateQuadricPixelRegion(actualPin, size);
                Boolean[] pixelRelevanceInformation = pinPixels.getRelevanceInformation(pixels,expression);
                Object[] pixelData = {pixels, pixelRelevanceInformation};
                outputData.put(pins[i], pixelData);
            }
        }
        return outputData;
    }

    /**
     * Calculates the number of pixels stored in the hashMap.
     * 
     * @param the hashmap providing the pixel coordinates
     * 
     * @return the number of pixels stored in the hashMap
     */
    private int getNumPixelsToExport(final HashMap pinPixels) {
        int numPixels = 0;
        Iterator iter = pinPixels.entrySet().iterator();
        while (iter.hasNext()) {
            Entry entry = (Entry) iter.next();
            Object pixelDataObject = entry.getValue();
            if (pixelDataObject != null && pixelDataObject instanceof Point[]) {
                Point[] pixels = (Point[]) entry.getValue();
                if (pixels != null) {
                    numPixels += pixels.length;
                }                
            } else if (pixelDataObject != null) {
                Object[] pixelData = (Object[]) entry.getValue();
                Point[] pixels = (Point[]) pixelData[0];
                if (pixels != null) {
                    numPixels += pixels.length;
                }     
            }
        }
        return numPixels;

    }

    /**
     * Creates the default filename for the exported data.
     */
    private String createDefaultFileName(final Product product) {
        return FileUtils.getFilenameWithoutExtension(product.getName()) + "_PinPixels.txt";
    }

    private String getWindowTitle() {
        return _visatApp.getAppName() + " - " + COMMAND_NAME;
    }
}